
/****************************************************************************/
//                           ***   ZXTAPE.C   ***

//             TRANSFER P FILES FROM PC PRINTER PORT TO ZX81
//      hardware and original BASIC software by W.Rigter (c) Aug 1996
//                 for information email rigter@cafe.net

//    Ported from BASIC to C by Roger DiPaolo 12/98 (dipaolor@eudoramail.com)
//    C compiler used is Borland C/C++ 3.1

//    The following LPT signals are used:
//    LPT data register (ioport0) - write data to data port (D0-D7).
//    Pin 2 - DATA BIT 0

//  version 2.0	12/12/98 rld	Initial release of 'C' version

//  version 2.1	12/19/98 rld	Fixed file listing to use all of screen
//				instead of scrolling one file name per
//				line.

//				Added loading of actual filename to ZX81.

//				Calibration is no longer run before each
//				load, to save time.  It is only run at the
//				start of the program.  There should be no
//				reason that timing should change after the
//				initial calibration.

//  version 2.2	12/20/98 rld	Added ability to select any I/O port address


/****************************************************************************/

#include "stdio.h"
#include "conio.h"
#include "dos.h"
#include "string.h"
#include "dir.h"
#include "process.h"
#include "ctype.h"
#include "time.h"
#include "alloc.h"
#include "io.h"
#include "math.h"

// Definitions

#define SUCCESS 1
#define FAILURE 0
#define YES 1           
#define NO 0
#define ON 1
#define OFF 0
#define OK 1
#define NOT_OK 0
#define QUIT 0
#define CONTINUE 1

#define ESCAPE 0x1B     // Escape
#define CTRL_LEFTBRACKET ESCAPE
#define CTRL_E 5		// Enquiry
#define ENQUIRY CTRL_E
#define CTRL_X 0x18		// Cancel
#define CANCEL CTRL_X
#define BELL 7          // ring bell through tty print
#define ZERONULL 0
#define BACKSPACE 8		// Backspace
#define SPACE 0x20		// Space
#define CARRIAGE_RETURN 0x0d

#define ECHO_ON 1
#define ECHO_OFF 0

#define BAD_CHOICE 0

#define BAD_DATA NOT_OK

#define MAXLINELEN 80
#define MAXDIRLEN 255

#define FUDGE_FACTOR 1.68  // Maybe I should be ashamed, but it works ;)

enum BOOL_ENUM {TRUE = 1, FALSE = 0};

typedef enum BOOL_ENUM BOOL;

// Prototypes

void getkey(void);
void dir1(void);
void getfile(void);
unsigned long calibrate_time(void);
void do_calibration(void);
unsigned GetHexValue(char *prompt_string, unsigned *status, BOOL NewLine);
int GetStrN(char *buffer,int echo,int maxchars);
void SetIOAddress(void);

// Globals

unsigned char 		*a = 0;
char			filestring[MAXLINELEN+1];
char			dirstring[MAXDIRLEN+1];
char			dir1string[MAXDIRLEN+1];
char			program_name[MAXLINELEN+1];
char			address_string[MAXLINELEN+1];
char			bytestring[MAXLINELEN+1];
char			inchar;
char			*dirptr;
int			ioport0, bytes, p;
unsigned int		n, address, timeout, looptime;
unsigned char		byte;
FILE			*infile;
int			tempword;
unsigned long		time_loops_per_ms;
unsigned long		pulse_hi_delay, pulse_low_delay, bit_delay;
float			ff = FUDGE_FACTOR;
struct text_info	ti;
int			tempchar;

/****************************************************************************/

main()
{
	unsigned long	i;
	unsigned char	maskbit = 1, b;
	unsigned int	l;
	char		dummy;


	clrscr();

	printf("\n\n  ZXTAPE v2.2 - TRANSFER P FILES IN TAPE FORMAT FROM PC TO ZX81 \n");
	printf("\nUse a 3.5mm headphone-stype plug connected to a male DB25 connector");
	printf("\n  with a two wires from tip to pin 2 and ring (ground) to pin 24\n");
	printf("\nHardware/software concept and original ZXTAPE by W.Rigter (c) Aug 1996");
	printf("\n                       email: rigter@cafe.net\n");
	printf("\n     Source code ported to 'C' from BASIC and modifications");
	printf("\n   Dec 1998 by Roger DiPaolo   email: dipaolor@eudoramail.com\n");

	getkey();

// set default values

	strcpy(filestring,"p.p");
	ioport0 = 0x378;
	dirptr = getcwd(dirstring,MAXLINELEN);

// Perform timing loop calibration

	do_calibration();

mainloop:

// initialize

	outp(ioport0,0);
	gettextinfo(&ti);

//    print menu

cont:

	clrscr();

	printf("\n Select one of the following ");

	printf("\n");
	printf("\n L) LOAD       ( Load a file to the ZX81 )");
	printf("\n F) FILE LIST  ( List P files in current directory )");
	printf("\n C) CHANGE DIR ( Current directory is %s )",dirstring);
	printf("\n B) BATCH      ( Call ZXTAPE.BAT )");
	printf("\n D) DOS        ( Type EXIT to return to ZXTAPE)");
	printf("\n I) IOPORT     ( Setup IOPORT )");
	printf("\n H) HELP       ( About ZXTAPE )");
	printf("\n Q) QUIT       ( Return to DOS )");
	printf("\n");

	getkey();

	if(inchar == 'l')
	{
		goto load;
	}

	if(inchar == 'f')
	{
		dir1();
		getkey();
	}

	if(inchar == 'b')
	{
		execlp("command.com","zxtape.bat",NULL);
		chdir(dirstring);
	}

	if(inchar == 'd')
	{
		printf("\nType EXIT to return from DOS");
		execlp("command.com",NULL);
	}

	if(inchar == 'i')
	{
		goto ioport;
	}

	if(inchar == 'c')
	{
		printf("\nEnter new working directory: ");
		gets(dir1string);
		if(!dir1string[0])
		{
			goto end;
		}
		else
		{
			strncpy(dirstring,dir1string,MAXDIRLEN);
		}
		chdir(dirstring);
	}
	
	if(inchar == 'h')
	{
		clrscr();

//		printf("\nHARDWARE:");
		printf("\nTurn your PC into a cassette player loading ZX81 programs with ZXTAPE");
		printf("\nJust make up a simple cable with 2 wires and a male DB25 connector");
		printf("\n(for the PC printer port) to one end and a 3.5mm headphone-style plug");
		printf("\n(for the ZX81 EAR jack) on the other end. Connect one wire to DB25");
		printf("\npin 2 and TIP terminal of a 3.5mm headphone-style plug. Connect the");
		printf("\nother wire to DB25 pin 24 and the RING (or ground) terminal of the\nplug.  That's it!");

//		printf("\nSOFTWARE:");
		printf("\nThe serial data generated by ZXTAPE simulates the normal tape signal");
		printf("\nrecorded by a ZX81. The pulses are generated with delays based on ");
		printf("\nsoftware loops which are calibrated at program initialization.  This should");
		printf("\ninsure accurate timing regardless of the speed of your PC's CPU.");

		printf("\n***** COMMAND SUMMARY *****");
		printf("\nL - On the PC type in a file name that you wish to load from the P file\n    listing, press enter, then follow instructions.");
		printf("\nF - Lists all the .P files in the current directory");
		printf("\nC - enter new directory : for example \\zx81\\files");
		printf("\nB - run batch file ZXTAPE.BAT in the current directory:");
		printf("\n    for example create a simple ASCII file with the text");
		printf("\n");
		printf("\n    XTenders");
		printf("\n");
		printf("\n    to run XTender from within the ZXTAPE program.");
		printf("\nD - Shell out to DOS and return by typing EXIT");
		printf("\nI - Select IOPORT address for your parallel port");

		printf("\n--> HIT ANY KEY TO RETURN TO MENU...");

		while(!kbhit());
	}

	if(inchar == 'q')
	{
		clrscr();
		exit(1);
	}

	goto mainloop;


load:
	getfile();
	if(bytes == 0)
	{
		printf("\nError attempting to load file");
		delay(3000);
		clrscr();
		goto mainloop;
	}


	clrscr();

outfile:

	printf("\nOn your ZX81 enter command: LOAD \"%s\" ,then hit any key on PC",program_name);
	while(!kbhit());
	dummy = getch();
	printf("\r                                                                                 ");
	printf("\rSending 5 seconds of silence...");
	outp(ioport0,0);
	delay(5000);  // lead in with 5 seconds of silence
	printf("\rTransfer will take about %d seconds",(int)ceil((18.0 * ((float)bytes)) / 1000.0));
	printf("\n*** If you wish to quit transfer, press space bar to abort ***");
	printf("\nLoading %d bytes of file %s",bytes,filestring);

	// Remove any "stray" keyhit from the key buffer that may cause
	// 'kbhit()' to take more processing time than usual, therefore
	// slowing down the main loop which would result in incorrect
	// byte xfer timing.

	if(kbhit())
	{
		tempchar = getch();
	}

	// This is the main xfer loop

	for(n=0;n<bytes;n++)  // Once through this loop for each byte
	{

		if(kbhit())  // Check for user abort here
		{
			if(getch() == ' ')
			{
				goto breakout;
			}
		}

		maskbit = 0x80;  // Start each byte by sending msb

		do  // Once through this loop for each bit
		{
			if(!(a[n] & maskbit))  // Determine if bit is 0 or 1
			{
			   p=4;  // A zero is 4 pulses
			}
			else
			{
			   p=9;  // A one is 9 pulses
			}

			for(l=0;l<p;l++)  // Output either 4 or 9 pulses
			{
				outp(ioport0,1);
				for(i=0;i<pulse_hi_delay;i++);	// high time is about .15ms
				outp(ioport0,0);
				for(i=0;i<pulse_low_delay;i++);	// low time is about .15ms
			}

			// Delay between bits

			for(i=0;i<bit_delay;i++);  // delay about 1.5ms before next bit

			maskbit >>= 1;  // Do next least signifigant bit

		}while(maskbit > 0);
	}

outfile1:

	printf(" - loading finished! ");
	goto end;

breakout:

	printf(" - load interrupted!");
	goto end;

ioport:

	SetIOAddress();

end:

	if(a)
	{
		free(a);
		a = (unsigned char *)0;
	}
	fclose(infile);
	printf("\n ");
	printf("\nWait 5 seconds, or press any key to continue...");

	timeout = 0;

	while(!kbhit())
	{
		delay(10);

		if(timeout++ == 500)  // Wait 5 seconds for a keyhit before 
		{		      // going automatically back
			break;	      // to mainloop
		}

	}

	goto mainloop;

}

/****************************************************************************/

void
getkey()
{
	inchar = tolower(getche());
}

/****************************************************************************/

void
dir1()
{  
	int 			done, line=2, column=1;
	int			dummy;
	struct find_t		ffblk;

	clrscr();
	printf("List of .P files in current directory :");
	done = _dos_findfirst("*.p",_A_NORMAL,&ffblk);

	while(!done)
	{
		gotoxy(column,line);
		printf("%s",ffblk.name);
		done = _dos_findnext(&ffblk);
		if(column <= 60)
		{
			column += 20;
		}
		else
		{
			line++;
			if(line == (ti.screenheight - 1))
			{
				gotoxy(1,ti.screenheight);
				printf("***** HIT ENTER KEY TO DISPLAY MORE FILES *****");
				dummy = getchar();
				clrscr();
				line = 1;
			}
			column = 1;
		}
	}
}

/****************************************************************************/

void
getfile()
{
	long	flen;
	int	program_name_len, i;


	clrscr();

	// Let user choose file

	dir1();
	printf("\nEnter name of data file without extension: ");
	gets(program_name);

	if(program_name != "")
	{
	   strncpy(filestring,program_name,MAXLINELEN);
	   strcat(filestring,".p");
	}
	else
	{
		return;
	}

	// Open file
	
	if(!(infile = fopen(filestring,"rb")))
	{
		sound(440);
		delay(500);
		nosound();
		printf("\nFile not found");
		return;
	}

	// Measure file length (in bytes)

	fseek(infile,0,SEEK_END);
	flen = ftell(infile);
	fseek(infile,0,SEEK_SET);

	program_name_len = strlen(program_name);

	// Allocate a buffer large enough to hold entire file

	a = (unsigned char *)malloc(flen+program_name_len+2);

	if(!a)
	{
		printf("\nNot enough memory to load file");
		return;
	}

	// Convert ASCII file name to "Sinclair Code" program name and store

	for(i=0;i<program_name_len;i++)
	{
		program_name[i] = toupper(program_name[i]);
		a[i] = program_name[i] - 27;
	}

	a[i-1] += 128;  // Make last character of program name "inverse"

	// Read file and load into buffer, 1 byte at a time

	for(bytes=i;;bytes++)
	{
		tempword = fgetc(infile);
		if(tempword == EOF)
		{
			break;
		}
		else
		{

			a[bytes] = (unsigned char)tempword;
		}
	}

}

/****************************************************************************/

unsigned long
calibrate_time()
{
	unsigned long	i, loops = 50000000;
	unsigned	start, end;
	unsigned long	IterationsPerMillisecond;

	printf("\nPerforming time calibration, please wait....");

	start = (unsigned)clock();
	for(i=0;i<loops;i++);
	end = (unsigned)clock();
	looptime = end - start;  // number of 55ms time chunks in loop
	IterationsPerMillisecond = (unsigned long)(loops / (unsigned long)(looptime*55));

	clrscr();

	return(IterationsPerMillisecond);

}

/****************************************************************************/

void
do_calibration(void)
{
	time_loops_per_ms = calibrate_time();

	bit_delay = (unsigned long)((float)time_loops_per_ms * 1.2 * ff);
	pulse_hi_delay = (unsigned long)((float)time_loops_per_ms * 0.13 * ff);
	pulse_low_delay = (unsigned long)((float)time_loops_per_ms * 0.11 * ff);
}

/*****************************************************************************
** NAME:	GetHexValue
**
** PURPOSE:		Retrieve numerical user input in hexadecimal format			
**
** PROTOTYPE:	unsigned GetHexValue(char *prompt_string, unsigned *status, BOOL NewLine)	
**
** ARGUMENTS:	
**		char		*prompt_string		-  pointer to string to prompt user
**		unsigned	*status				-  a pointer to an unsigned word  
**											  
**			The following data is stored in *status for processing by calling
**			function:
**					ESCAPE				- escape key was pressed.
**					BAD_DATA			- no data entered or data not in hex
**										format.
**					OK 					- none of the previously described 
**										events occurred.
**
**		BOOL		NewLine			- if TRUE insert a newline before the prompt.
**
** RETURN VALUE:	Returns hexadecimal value entered OR returns a 0
**						for the following conditions:
**						1.  Escape key was pressed.
**						2.  No characters or more than 8 characters entered when	
**							prompted for input.
**						3.  Data entered is not in hexadecimal format. 
**   
** GLOBALS USED:	None	
**
** REMARKS:	
**
******************************************************************************/

unsigned
GetHexValue(char *prompt_string, unsigned *status, BOOL NewLine)
{
	char		buffer[11];
	int			num_of_chars;
	int			GetStrStatus;
	unsigned	value;

	*status = OK;

	if(NewLine == TRUE)
	{
		printf("\n%s",prompt_string);
	}
	else
	{
		printf("%s",prompt_string);
	}

	GetStrStatus = GetStrN(buffer,ECHO_ON,8);
	
	if(GetStrStatus == ESCAPE)
	{
		*status = ESCAPE;
		return(0);										
	}

	buffer[10] = '\0';

	num_of_chars = strlen(buffer);

	if((num_of_chars > 8) || (num_of_chars < 1))
	{
		*status = BAD_DATA;								
		return(0);										
	}	

	if(!sscanf(buffer,"%lx",&value))
	{

		// Error: data entered not in hexadecimal format.
		// Note: data does not have to be preceded by hex value delimiter '0x'
		*status = BAD_DATA;							   
		return(0);										
	}

	return(value);
}


/****************************************************************************/

int
GetStrN(char *buffer,int echo,int maxchars)
{
	int		charcount = 0;

	while(1)
	{
		buffer[charcount] = getch();
		switch(buffer[charcount])
		{
			case ESCAPE:
							putchar('\n');
							buffer[0] = '\0';
							return(ESCAPE);

			case BACKSPACE:
							if(charcount > 0)
							{
								if(echo == ECHO_ON)
								{
									putchar(BACKSPACE);
									putchar(SPACE);
									putchar(BACKSPACE);
								}
								charcount--;
							}
							break;

			case CARRIAGE_RETURN:
							putchar(0x0d);
							putchar(0x0a);
							buffer[charcount] = '\0';
							return(0);

			default:
							if((buffer[charcount] >= SPACE) && (buffer[charcount] <= 0x7e)) 
							{
								if(charcount < maxchars)
								{
									if(echo == ECHO_ON)
									{
										putchar(buffer[charcount]);
									}
									charcount++;
								}
							}
		}
	}
}

/****************************************************************************/

void
SetIOAddress(void)
{
	int		newportaddr;
	unsigned	status;

	clrscr();
	address = ioport0;
	printf("\nThe current I/O port address is %X hex",address);
	printf("\nSelect one of the following choices:");
	printf("\n1 = Use port at address 278 hex");
	printf("\n2 = Use port at address 378 hex");
	printf("\n3 = Enter an address not shown above");
	printf("\n\n>");
	getkey();

	if(inchar == '1')
	{
		ioport0 = 0x278;
	}
	else if(inchar == '2')
	{
		ioport0 = 0x378;
	}
	else if(inchar == '3')
	{
		if((newportaddr = GetHexValue("Enter I/O port address (in hex): ",&status,TRUE)) == 0)
		{
			if(status == ESCAPE)
			{
				printf("\n>> Enter Aborted <<");
			}
			else if(status == BAD_DATA)
			{
				printf("\n>> Invalid data, Address unchanged <<");
			}
		}
		else
		{
			ioport0 = newportaddr;
		}
	}

	printf("\nI/O port address is %X hex",ioport0);
}
